.termynal-comment {
    color: #4a968f;
    font-style: italic;
    display: block;
}

.termy [data-termynal] {
    white-space: pre-wrap;
    font-size: 14px;
    overflow-wrap: break-word;
}

a.external-link::after {
    /* \00A0 is a non-breaking space
        to make the mark be on the same line as the link
    */
    content: "\00A0[↪]";
}

a.internal-link::after {
    /* \00A0 is a non-breaking space
        to make the mark be on the same line as the link
    */
    content: "\00A0↪";
}